# Copyright (c) 2024-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from neuzip.ops.linear import CompressedLinear

__all__ = ["CompressedLinear"]
